# This file is part of HYLOA - HYsteresis LOop Analyzer.
# Copyright (C) 2024 Francesco Zeno Costanzo

# HYLOA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# HYLOA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with HYLOA. If not, see <https://www.gnu.org/licenses/>.

"""
Code with basic routines to correct some loop distortions
"""

import numpy as np
from scipy.optimize import curve_fit
from PyQt5.QtWidgets import QMessageBox

from hyloa.utils.err_format import format_value_error

#================================================#
# Function to chek simmetry by flipping a branch #
#================================================#

def change_ps(plot_state, window, draw_plot, mode="cp"):
    '''
    Function to change the plot status deleting the corrected
    or the original data

    Parameters
    ----------
    plot_state : dict
        Dictionary storing the current plotting state.
    window : QWidget
        Parent widget used to display error message boxes.
    draw_plot : callable
        Callback function responsible for redrawing the plot.
    mode : string, optional, dafult "cp"
        if mode="cp" all correction will be remove from the plot
        id mode="od" the original data will be removed from the plot
        if mode="sym" the symmetrized data will be removed from the plot
    '''

    try:
        if mode =="cp":
            plot_state.update({
                "done_corr": False,
                "done_spl3": False,
                "x_up_corr": None,
                "y_up_corr": None,
                "e_up"     : None,
                "x_dw_corr": None,
                "y_dw_corr": None,
                "e_dw"     : None,
                "fit_hc_p" : None,
                "fit_hc_n" : None,
                "spline_up": None,
                "spline_dw": None
            })
            draw_plot()

        if mode == "od":
            plot_state.update({
                "x_up" : None,
                "y_up" : None,
                "x_dw" : None,
                "y_dw" : None,
            })
            draw_plot()
        
        if mode == "sym":
            plot_state.update({
                "s_data_up" : None,
                "s_data_dw" : None
            })
            draw_plot()

    except Exception as e:
        QMessageBox.critical(window, "Error", f"Error during flip:\n{e}")

#================================================#
# Function to chek simmetry by flipping a branch #
#================================================#

def flip(plot_state, window, draw_plot):
    '''
    Function to flip a brach to ensure simmetricity of the loop.

    Parameters
    ----------
    plot_state : dict
        dictionary of the plotted data
    window : QWidget
        The main window widget.
    draw_plot : callable
        Function to update the preview
    '''
    try:
        plot_state["flipped"] = not plot_state["flipped"]
        draw_plot()

    except Exception as e:
        QMessageBox.critical(window, "Error", f"Error during flip:\n{e}")

#================================================#
# Function to flip a selected branch             #
#================================================#

def flip_data(file_combo,
              x_up_combo, y_up_combo, x_down_combo, y_down_combo,
              data_sel, double_branch, plot_state,
              window, logger, draw_plot):
    '''
    Function to duplicate a branch by flipping it to recontruct a simmetric loop.

    Parameters
    ----------
    file_combo : QComboBox
        Combo box to select source data file.
    x_up_combo : QComboBox
        Combo box to select X column for Up branch.
    y_up_combo : QComboBox
        Combo box to select Y column for Up branch.
    x_down_combo : QComboBox
        Combo box to select X column for Down branch.
    y_down_combo : QComboBox
        Combo box to select Y column for Down branch.
    data_sel : QComboBox
        Combo box to select data type (corrected or original).
    double_branch : QComboBox
        Combo box to select a branch duplication.
    plot_state : dict
        Dictionary storing the current plotting state.
    window : QWidget
        Parent widget used to display error message boxes.
    logger : logging.Logger
        Logger instance used to record spline computation details.
    draw_plot : callable
        Callback function responsible for redrawing the plot.
    '''
    try:
        idx_src  = file_combo.currentIndex()
        selected = data_sel.currentText()
        x_up_col = x_up_combo.currentText()
        y_up_col = y_up_combo.currentText()
        x_dw_col = x_down_combo.currentText()
        y_dw_col = y_down_combo.currentText()

        #Read data
        if selected == "Corrected":
            x_up = plot_state["x_up_corr"]
            y_up = plot_state["y_up_corr"]
            x_dw = plot_state["x_dw_corr"]
            y_dw = plot_state["y_dw_corr"]
            e_up = plot_state["e_up"]
            e_dw = plot_state["e_dw"]
        else:
            x_up = plot_state["x_up"]
            y_up = plot_state["y_up"]
            x_dw = plot_state["x_dw"]
            y_dw = plot_state["y_dw"]
            tail = np.concatenate((y_up[0:25], y_dw[-25:], 
                                   y_up[-25:], y_dw[0:25]))
            dy_data_err = np.std(tail)
            dy_err = (2*np.random.random(x_up.size) - 1) * dy_data_err
            e_up = dy_err
            e_dw = dy_err

        if x_up is None:
            QMessageBox.critical(window, "Error", "This can be done only on corrected data.")
            return

        db = double_branch.currentText()

        if db == "No":
            # Do nothing if button will be pressed
            return

        elif db == "Up":
            x_dw, y_dw = -np.copy(x_up), -np.copy(y_up)
            x_dw, y_dw = x_dw[::-1], y_dw[::-1]  # reverse to maintain order
            e_dw       = np.copy(e_up)[::-1]

        elif db == "Down":
            x_up, y_up = -np.copy(x_dw), -np.copy(y_dw)
            x_up, y_up = x_up[::-1], y_up[::-1]  # reverse to maintain order
            e_up       = np.copy(e_dw)[::-1]

        plot_state.update({
            "x_up_corr": x_up,
            "y_up_corr": y_up,
            "e_up"     : e_up,
            "x_dw_corr": x_dw,
            "y_dw_corr": y_dw,
            "e_dw"     : e_dw,
        })
        draw_plot()

        log_lines = []
        log_lines.append(f"Flipped {db} branch in file {idx_src + 1}, columns {x_up_col}/{y_up_col} and {x_dw_col}/{y_dw_col}.")
        
        logger.info("\n".join(log_lines))

    
    except Exception as e:
        QMessageBox.critical(window, "Error", f"Error during branch flipping:\n{e}")


#================================================#
# Function to correct field                      #
#================================================#

def apply_shift(x_up_dest, y_up_dest, x_dw_dest, y_dw_dest, dataframes, save_file_combo,
                data_sel, field_shift_pc_edit, plot_state, window, fit_data, args=(),
                logger=None):
    '''
    Function add a field shift after the corrections

    Parameters
    ----------
    x_up_dest : QComboBox
        Combo box to select X column in which to store the correct up branch.
    y_up_dest : QComboBox
        Combo box to select Y column in which to store the correct up branch.
    x_dw_dest : QComboBox
        Combo box to select X column in which to store the correct down branch.
    y_dw_dest : QComboBox
        Combo box to select Y column in which to store the correct down branch.
    dataframes : list of pd.DataFrame
        List of dataframes containing loaded data.
    save_file_combo : QComboBox
        Combo box to select destination data file.
    field_shift_pc_edit : QLineEdit
        Value for field shifting
    plot_state : dict
        Dictionary storing the current plotting state.
    window : QWidget
        Parent widget used to display error message boxes.
    fit_data : callable
        Function for fitting data
    args : tuple
        Argumets to pass to fit_data
    logger : logging.Logger
        Logger instance used to record spline computation details.
    '''

    save_choice  = save_file_combo.currentIndex()  # 0 = No save, >0 => file index adjust
    selected     = data_sel.currentText()
        
    if save_choice == 0:
        save_idx = None
    else:
        # save_file_combo items: ["No save", "File 1", "File 2"...]
        save_idx = save_choice - 1

    #Read data
    if selected == "Corrected":
        name_xup = "x_up_corr"
        name_xdw = "x_dw_corr"
        name_yup = "y_up_corr"
        name_ydw = "y_dw_corr"
  
    else:
        name_xup = "x_up"
        name_xdw = "x_dw"
        name_yup = "y_up"
        name_ydw = "y_dw"

    try:

        field_shift = float(field_shift_pc_edit.text())
        plot_state[name_xup] -= field_shift
        plot_state[name_xdw] -= field_shift

        try :
            fit_data(*args)
            if save_idx is not None:
                df_dest = dataframes[save_idx]

                df_dest[x_up_dest.currentText()] = plot_state[name_xup]
                df_dest[x_dw_dest.currentText()] = plot_state[name_xdw]
                df_dest[y_up_dest.currentText()] = plot_state[name_yup]
                df_dest[y_dw_dest.currentText()] = plot_state[name_ydw]


                logger.info("Corrected columns saved to destination file.")

        except Exception as e:
            QMessageBox.critical(window, "Error", f"Error during fit:\n{e}")
            # Return to original values
            plot_state[name_xup] += field_shift
            plot_state[name_xdw] += field_shift

    except Exception as e:
        QMessageBox.critical(window, "Error", f"Error applying shift:\n{e}")


#================================================#
# MAIN CORRECTION FUNCTION                       #
#================================================#

def perform_correction(file_combo, save_file_combo,
                       x_up_combo, y_up_combo, x_down_combo, y_down_combo,
                       field_shift_edit, field_scale_edit,
                       x_start_n_edit, x_end_n_edit, x_start_p_edit, x_end_p_edit,
                       tail_params_edit, tail_function_edit,
                       x_up_dest, y_up_dest, x_dw_dest, y_dw_dest,
                       smooth_up_edit, smooth_dw_edit,
                       dataframes, logger, plot_state, draw_plot,
                       window):
    ''' 
    Perform the loop correction using the parameters from the UI.
    The correction involves fitting the saturation parts (tails) of the hystersis loop
    to remove drifts.
    Using the computed error the code also provides an estimation of
    smoothing parameters for Bspline fitting.

    Parameters
    ----------
    file_combo : QComboBox
        Combo box to select source data file.
    save_file_combo : QComboBox
        Combo box to select destination data file.
    x_up_combo : QComboBox
        Combo box to select X column for Up branch.
    y_up_combo : QComboBox
        Combo box to select Y column for Up branch.
    x_down_combo : QComboBox
        Combo box to select X column for Down branch.
    y_down_combo : QComboBox
        Combo box to select Y column for Down branch.
    field_shift_edit : QLineEdit
        Line edit for field shift value.
    field_scale_edit : QLineEdit
        Line edit for field scale value.
    x_start_n_edit : QLineEdit
        Line edit for negative tail start limit.
    x_end_n_edit : QLineEdit
        Line edit for negative tail end limit.
    x_start_p_edit : QLineEdit
        Line edit for positive tail start limit.
    x_end_p_edit : QLineEdit
        Line edit for positive tail end limit.
    tail_params_edit : QLineEdit
        Line edit for tail fit parameter names.
    tail_function_edit : QLineEdit
        Line edit for tail fit function.
    x_up_dest : QComboBox
        Combo box to select X column in which to store the correct up branch.
    y_up_dest : QComboBox
        Combo box to select Y column in which to store the correct up branch.
    x_dw_dest : QComboBox
        Combo box to select X column in which to store the correct down branch.
    y_dw_dest : QComboBox
        Combo box to select Y column in which to store the correct down branch.
    dataframes : list of pd.DataFrame
        List of dataframes containing loaded data.
    smooth_up_edit : QLineEdit
        Line edit for smoothing parameter for spline for up branch
    smooth_dw_edit : QlineEdit
        Line edit for smoothing parameter for spline for down branch
    logger : logging.Logger
        Logger instance used to record spline computation details.
    plot_state : dict
        Dictionary storing the current plotting state.
    draw_plot : callable    
        Callback function responsible for redrawing the plot.
    output_box : QTextEdit
        Text edit for displaying output results.
    window : QWidget
        Parent widget used to display error message boxes.
    '''
    try:
        idx_src      = file_combo.currentIndex()
        df_src       = dataframes[idx_src].copy()      # working copy
        save_choice  = save_file_combo.currentIndex()  # 0 = No save, >0 => file index adjust
        
        if save_choice == 0:
            save_idx = None
        else:
            # save_file_combo items: ["No save", "File 1", "File 2"...]
            save_idx = save_choice - 1

        # Read field shift/scale
        field_shift = float(field_shift_edit.text())
        field_scale = float(field_scale_edit.text())

        # Read x/y column names
        x_up_col = x_up_combo.currentText()
        y_up_col = y_up_combo.currentText()
        x_dw_col = x_down_combo.currentText()
        y_dw_col = y_down_combo.currentText()
        
        # Prepare arrays: apply shift and scale
        x_up = np.copy(df_src[x_up_col].astype(float).values) - field_shift
        x_up = x_up * field_scale
        y_up = np.copy(df_src[y_up_col].astype(float).values)
        x_dw = np.copy(df_src[x_dw_col].astype(float).values) - field_shift
        x_dw = x_dw * field_scale
        y_dw = np.copy(df_src[y_dw_col].astype(float).values)

        # Read tail ranges
        try:
            x_n_start = float(x_start_n_edit.text()) * field_scale # For up negative region
            x_n_end   = float(x_end_n_edit.text()) * field_scale
        except Exception:
            # fallback
            x_n_start, x_n_end = -4000,  -1000

        try:
            x_p_start = float(x_start_p_edit.text()) * field_scale # For up positive region
            x_p_end   = float(x_end_p_edit.text()) * field_scale 
        except Exception:
            x_p_start, x_p_end = 1000, 4000

        # Masks for up/down tails split by sign
        mask_n_up = (x_up >= x_n_start) & (x_up <= x_n_end)
        mask_p_up = (x_up >= x_p_start) & (x_up <= x_p_end)
        mask_n_dw = (x_dw >= x_n_start) & (x_dw <= x_n_end)
        mask_p_dw = (x_dw >= x_p_start) & (x_dw <= x_p_end)
        
        #=========================================#
        # Auxiliary functions for fitting         #
        #=========================================#

        try:
            tail_param_names = [p.strip() for p in tail_params_edit.text().split(",") if p.strip() != ""]
            func_code_tail   = f"lambda x, {', '.join(tail_param_names)}: {tail_function_edit.text()}"
            f_func           = eval(func_code_tail, {"np": np, "__builtins__": {}})
        
        except Exception as e:
            QMessageBox.critical(window, "Error", f"Invalid function for tail fit:\n{e}")
            return
        
        s2 = lambda x, y, f, popt: sum((y-f(x, *popt))**2)/(len(x)-len(popt))

        def poly_error(x, params, dparams):
            '''
            params: [a0, a1, a2, ...]
            dparams: relative uncertainties
            '''
            total = 0.0
            # Skip i=0 because a0 is the constant term
            for i in range(1, len(params)):
                total += ( (x**i) * dparams[i] )**2
            return np.sqrt(total)

        #=========================================#

        # Fit linear tails (four fits)
        p_up_1, c_up_1 = curve_fit(f_func, x_up[mask_n_up], y_up[mask_n_up])
        p_up_2, c_up_2 = curve_fit(f_func, x_up[mask_p_up], y_up[mask_p_up])
        p_dw_1, c_dw_1 = curve_fit(f_func, x_dw[mask_n_dw], y_dw[mask_n_dw])
        p_dw_2, c_dw_2 = curve_fit(f_func, x_dw[mask_p_dw], y_dw[mask_p_dw])

        # Parameter errors
        dp_up_1 = np.sqrt(np.diag(c_up_1))
        dp_up_2 = np.sqrt(np.diag(c_up_2))
        dp_dw_1 = np.sqrt(np.diag(c_dw_1))
        dp_dw_2 = np.sqrt(np.diag(c_dw_2))

        # Model dispersion error
        e_up_1 = s2(x_up[mask_n_up], y_up[mask_n_up], f_func, p_up_1)
        e_up_2 = s2(x_up[mask_p_up], y_up[mask_p_up], f_func, p_up_2)
        e_dw_1 = s2(x_dw[mask_n_dw], y_dw[mask_n_dw], f_func, p_dw_1)
        e_dw_2 = s2(x_dw[mask_p_dw], y_dw[mask_p_dw], f_func, p_dw_2)

        e_up = (e_up_1 + e_up_2) * 0.5
        e_dw = (e_dw_1 + e_dw_2) * 0.5

        # Corrections
        corr_up = np.where(
            x_up < 0,
            f_func(x_up, *p_up_1) - p_up_1[0],
            f_func(x_up, *p_up_2) - p_up_2[0]
        )

        corr_dw = np.where(
            x_dw < 0,
            f_func(x_dw, *p_dw_1) - p_dw_1[0],
            f_func(x_dw, *p_dw_2) - p_dw_2[0]
        )
        
        d_corr_up = np.where(
            x_up < 0,
            poly_error(x_up, p_up_1, dp_up_1),
            poly_error(x_up, p_up_2, dp_up_2)
        )

        d_corr_dw = np.where(
            x_dw < 0,
            poly_error(x_dw, p_dw_1, dp_dw_1),
            poly_error(x_dw, p_dw_2, dp_dw_2)
        )

        # Total error = dispersion + abs(d_corr)
        e_up = e_up + np.abs(d_corr_up)
        e_dw = e_dw + np.abs(d_corr_dw)

        # Update with an edicated guess
        smooth_up_edit.setText(f"{sum(e_up**2):.2f}")
        smooth_dw_edit.setText(f"{sum(e_dw**2):.2f}")


        # Apply correction to data
        y_up_corr = y_up - corr_up
        y_dw_corr = y_dw - corr_dw

        def close_fun(up_arr, q1, q2):
            ''' Close the loop using intercepts q1 and q2
            '''
            up_c  = np.copy(up_arr) - q1
            up_c *= 2/(abs(q2)+abs(q1))
            up_c  = up_c + q1/abs(q1) #- 1 if q1 < 0 else + 1
            return up_c

        # Apply close operation using intercepts q = intercept of negative and positive tail fits
        y_up_closed = close_fun(y_up_corr, p_up_1[0], p_up_2[0])
        y_dw_closed = close_fun(y_dw_corr, p_dw_1[0], p_dw_2[0])

        plot_state.update({
            "done_corr": True,
            "x_up_corr": x_up,
            "y_up_corr": y_up_closed,
            "e_up"     : e_up,
            "x_dw_corr": x_dw,
            "y_dw_corr": y_dw_closed,
            "e_dw"     : e_dw,
        })
        draw_plot()

        # Summary of results
        log_results_lines = []
        log_results_lines.append(f"Summary of correction for data in file {idx_src + 1}, columns {x_up_col}/{y_up_col} and {x_dw_col}/{y_dw_col}:")
        log_results_lines.append(f"Corrected data with field shift = {field_shift} and scale = {field_scale}")
        log_results_lines.append(f"Using tail fit function: {tail_function_edit.text()}")
        log_results_lines.append(f"Using range limits (neg): {x_n_start/field_scale} to {x_n_end/field_scale}")
        log_results_lines.append(f"Using range limits (pos): {x_p_start/field_scale} to {x_p_end/field_scale}\n")
        logger.info("Loop correction completed. Summary:\n" + "\n".join(log_results_lines))

        # Save corrected columns if requested
        if save_idx is not None:
            df_dest = dataframes[save_idx]

            df_dest[x_up_dest.currentText()] = x_up
            df_dest[y_up_dest.currentText()] = y_up_closed

            df_dest[x_dw_dest.currentText()] = x_dw
            df_dest[y_dw_dest.currentText()] = y_dw_closed

            logger.info("Corrected columns saved to destination file.")

    except Exception as e:
        QMessageBox.critical(window, "Error", f"Error running fits: {e}")
        logger.exception("Unhandled error in perform_all_fits: %s", e) 
    

#===================================================================#
# Function to fit for physiscal quantities                          #
#===================================================================#       

def fit_data(file_combo,
             x_up_combo, y_up_combo, x_down_combo, y_down_combo, data_sel,
             x_start_up_hc_edit, x_end_up_hc_edit, x_start_dw_hc_edit, x_end_dw_hc_edit,
             hc_params_edit, hc_function_edit, logger, plot_state, draw_plot,
             output_box, window):
    '''
    Parameters
    ----------
    file_combo : QComboBox
        Combo box to select source data file.
    x_up_combo : QComboBox
        Combo box to select X column for Up branch.
    y_up_combo : QComboBox
        Combo box to select Y column for Up branch.
    x_down_combo : QComboBox
        Combo box to select X column for Down branch.
    y_down_combo : QComboBox
        Combo box to select Y column for Down branch.
    data_sel : QComboBox
        Combo box to select data type (corrected or original).
    x_start_up_hc_edit : QLineEdit
        Line edit for Up branch coercive fit start limit.
    x_end_up_hc_edit : QLineEdit
        Line edit for Up branch coercive fit end limit.
    x_start_dw_hc_edit : QLineEdit
        Line edit for Down branch coercive fit start limit.
    x_end_dw_hc_edit : QLineEdit
        Line edit for Down branch coercive fit end limit.
    hc_params_edit : QLineEdit
        Line edit for coercive fit parameter names.
    hc_function_edit : QLineEdit
        Line edit for coercive fit function.
    logger : logging.Logger
        Logger instance used to record spline computation details.
    plot_state : dict
        Dictionary storing the current plotting state.
    draw_plot : callable    
        Callback function responsible for redrawing the plot.
    output_box : QTextEdit
        Text edit for displaying output results.
    window : QWidget
        Parent widget used to display error message boxes.
    '''
    try : 
        idx_src  = file_combo.currentIndex()
        selected = data_sel.currentText()
    
        # Read x/y column names
        x_up_col = x_up_combo.currentText()
        y_up_col = y_up_combo.currentText()
        x_dw_col = x_down_combo.currentText()
        y_dw_col = y_down_combo.currentText()

        #Read data
        if selected == "Corrected":
            x_up = plot_state["x_up_corr"]
            y_up = plot_state["y_up_corr"]
            x_dw = plot_state["x_dw_corr"]
            y_dw = plot_state["y_dw_corr"]
            e_up = plot_state["e_up"]
            e_dw = plot_state["e_dw"]
        else:
            x_up = plot_state["x_up"]
            y_up = plot_state["y_up"]
            x_dw = plot_state["x_dw"]
            y_dw = plot_state["y_dw"]

            tail = np.concatenate((y_up[0:25], y_dw[-25:], 
                                   y_up[-25:], y_dw[0:25]))
            dy_data_err = np.std(tail)
            dy_err = (2*np.random.random(x_up.size) - 1) * dy_data_err
            e_up = dy_err
            e_dw = dy_err
        
        if e_up is None:
            QMessageBox.critical(window, "Error", f"You need to correct data first")
            return

        
        results_text_lines =  []
        try:
            hc_param_names = [p.strip() for p in hc_params_edit.text().split(",") if p.strip() != ""]
            func_code_hc   = f"lambda x, {', '.join(hc_param_names)}: {hc_function_edit.text()}"
            g_func         = eval(func_code_hc, {"np": np, "__builtins__": {}})
        
        except Exception as e:
            QMessageBox.critical(window, "Error", f"Invalid function for coercive fit:\n{e}")
            return
        
        # Fit coercivity
        try:
            x_n_start_hc = float(x_start_dw_hc_edit.text())
            x_n_end_hc   = float(x_end_dw_hc_edit.text())
            x_p_start_hc = float(x_start_up_hc_edit.text())
            x_p_end_hc   = float(x_end_up_hc_edit.text())
        except Exception as e:
            QMessageBox.critical(window, "Error", f"Invalid value for Hc range:\n{e}")
            return
        
        mask_n = (x_dw >= x_n_start_hc) & (x_dw <= x_n_end_hc)
        mask_p = (x_up >= x_p_start_hc) & (x_up <= x_p_end_hc)

        if mask_n.sum() < 2 or mask_p.sum() < 2:
            QMessageBox.critical(window, "Error", "Not enough points in coercive fit ranges.")
            return
        else :
            # Perform weighted fits
            popt_n, covm_n = curve_fit(g_func, x_dw[mask_n], y_dw[mask_n], sigma=e_dw[mask_n])
            popt_p, covm_p = curve_fit(g_func, x_up[mask_p], y_up[mask_p], sigma=e_up[mask_p])

        
        t1 = np.linspace(x_n_start_hc, x_n_end_hc, 400)
        t2 = np.linspace(x_p_start_hc, x_p_end_hc, 400)

        plot_state.update({
            "fit_hc_p" : (t1, g_func(t1, *popt_n)),
            "fit_hc_n" : (t2, g_func(t2, *popt_p))
        })
        draw_plot()

        # Store numerical results
        results_text_lines.append("Coercive fit results:")
        for p, val, err in zip(hc_param_names, popt_n, np.sqrt(np.diag(covm_n))):
            try :
                results_text_lines.append(f"{p} = {format_value_error(val, err)}")
            except Exception as e:
                results_text_lines.append(f"{p} = {val:.6f} ± {err:.6f}")

        for i , pi in zip(range(len(popt_n)), hc_param_names):
            for j , pj in zip(range(i+1, len(popt_n)), hc_param_names[i+1:]):
                corr_ij = covm_n[i, j]/np.sqrt(covm_n[i, i]*covm_n[j, j])
                results_text_lines.append(f"corr({pi}, {pj}) = {corr_ij:.3f}")
        
        for p, val, err in zip(hc_param_names, popt_p, np.sqrt(np.diag(covm_p))):
            try :
                results_text_lines.append(f"{p} = {format_value_error(val, err)}")
            except Exception as e:
                results_text_lines.append(f"{p} = {val:.6f} ± {err:.6f}")    
            
        for i , pi in zip(range(len(popt_p)), hc_param_names):
            for j , pj in zip(range(i+1, len(popt_p)), hc_param_names[i+1:]):
                corr_ij = covm_p[i, j]/np.sqrt(covm_p[i, i]*covm_p[j, j])
                results_text_lines.append(f"corr({pi}, {pj}) = {corr_ij:.3f}")

        # Show textual results
        output_box.setPlainText("\n".join(results_text_lines))

        # Summary of results
        log_results_lines = []
        log_results_lines.append(f"Summary of fit for data in file {idx_src + 1}, columns {x_up_col}/{y_up_col} and {x_dw_col}/{y_dw_col}:")
        log_results_lines.append(f"Using fit function: {hc_function_edit.text()}")
        log_results_lines.append(f"Using fit ranges (down): {x_n_start_hc} to {x_n_end_hc}")
        log_results_lines.append(f"Using fit ranges (up): {x_p_start_hc} to {x_p_end_hc}\n")
        logger.info("Fit completed. Summary:\n" + "\n".join(log_results_lines) +"\n".join(results_text_lines))
    
    except Exception as e:
        QMessageBox.critical(window, "Error", f"Error during fitting:\n{e}")
        return




# This file is part of HYLOA - HYsteresis LOop Analyzer.
# Copyright (C) 2024 Francesco Zeno Costanzo

# HYLOA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# HYLOA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with HYLOA. If not, see <https://www.gnu.org/licenses/>.


"""
Entry point for HYLOA application.
"""

import sys
import time 
from importlib import resources
from PyQt5.QtWidgets import QApplication, QWidget, QVBoxLayout, QLabel, QProgressBar
from PyQt5.QtGui import QPixmap
from PyQt5.QtCore import Qt, QTimer


# Duration to show the splash screen at minimum
MIN_SPLASH_TIME = 3.0  # Seconds


class Splash(QWidget):
    '''
    Splash screen with a logo and a progress bar.
    '''
    def __init__(self, pixmap):
        '''
        Initialize the splash screen with a logo and a progress bar.

        Parameters
        ----------
        pixmap : QPixmap
            The logo to display on the splash screen.
        '''
        super().__init__(
            flags=Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint
        )

        self.setAttribute(Qt.WA_TranslucentBackground)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(20, 20, 20, 20)

        # Logo
        self.logo = QLabel()
        self.logo.setPixmap(pixmap)
        self.logo.setAlignment(Qt.AlignCenter)

        # Progress Bar
        self.progress = QProgressBar()
        self.progress.setRange(0, 100)
        self.progress.setValue(0)
        self.progress.setTextVisible(True)
        self.progress.setFormat("%p%")

        layout.addWidget(self.logo)
        layout.addWidget(self.progress)

        self.adjustSize()
        self.move(
            self.screen().geometry().center() - self.rect().center()
        )

    def set_progress(self, value):
        '''
        Update the progress bar value.
        
        Parameters
        ----------
        value : int
            The new value for the progress bar (0-100).
        '''
        self.progress.setValue(value)


def compute_remaining_time(start_time, min_splash_time):
    '''
    Compute the remaining time to show the splash screen

    Parameters
    ----------
    start_time : float
        The time when the splash screen was shown (monotonic).
    min_splash_time : float
        The minimum duration to show the splash screen (seconds).
    Returns
    -------
    int
        Remaining time in milliseconds to show the splash screen.
    '''
    elapsed = time.monotonic() - start_time
    return max(0, int((min_splash_time - elapsed) * 1000))


def main():
    '''
    Main entry point for the HYLOA application.
    '''
    start_time = time.monotonic()
    app = QApplication(sys.argv)

    # Load splash screen resources
    with resources.path("hyloa.resources", "icon-6.png") as p:
        pixmap = QPixmap(str(p))

    splash = Splash(pixmap)
    splash.show()
    splash.set_progress(20)
    app.processEvents()

    # Import main window (heavy part of the loading)
    from hyloa.gui.main_window import MainApp
    splash.set_progress(60)
    app.processEvents()
    window = MainApp()
   
    remaining = compute_remaining_time(start_time, MIN_SPLASH_TIME)

    def finish():
        splash.set_progress(100)
        app.processEvents()
        window.show()
        splash.close()

    QTimer.singleShot(remaining, finish)
    

    sys.exit(app.exec_())

if __name__ == "__main__":
    main()



# This file is part of HYLOA - HYsteresis LOop Analyzer.
# Copyright (C) 2024 Francesco Zeno Costanzo

# HYLOA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# HYLOA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with HYLOA. If not, see <https://www.gnu.org/licenses/>.


"""
Code to manage the plot window
"""
import numpy as np
from scipy.special import *
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from matplotlib.figure import Figure
from matplotlib import markers, lines as mlines, colors as mcolors
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5 import NavigationToolbar2QT as NavigationToolbar

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (
    QWidget, QVBoxLayout, QHBoxLayout, QLabel, QPushButton,
    QScrollArea, QComboBox, QMessageBox, QDialog, QFormLayout,
    QLineEdit, QMdiSubWindow, QTextEdit, QSizePolicy, QFrame,
    QCheckBox, QDialogButtonBox, QMenu, QAction, QGridLayout,
    QGroupBox, QStackedWidget, QButtonGroup

)

from hyloa.utils.err_format import format_value_error
from hyloa.data.processing import inv_single_branch_dialog
from hyloa.data.processing import inv_x_dialog, inv_y_dialog
from hyloa.data.processing import norm_dialog, close_loop_dialog
from hyloa.gui.correction_window import correct_hysteresis_loop

#==============================================================================================#
# Main class for managing the plot window                                                      #
#==============================================================================================#

class PlotControlWidget(QWidget):
    '''
    Main widget for managing a single plot's interface and user interactions.

    Allows the user to configure and customize data visualization options, including
    selecting datasets, customizing styles, fitting curves, and toggling data visibility.
    '''

    def __init__(self, app_instance, number_plots, plot_name="Graph"):
        '''
        Initialize the plot control widget and build the GUI layout.
        
        Parameters
        ----------
        app_instance : object
            The main application instance that contains shared data and methods.
        number_plots : int
            Index or identifier of the current plot.
        plot_name : str, optional
            The display name of the plot window (default is "Graph").
        '''
        super().__init__()
        
        self.app_instance         = app_instance  # Instance of main app
        self.number_plots         = number_plots  # Index of the plot
        self.plot_customizations  = {}            # Dictionary to save graphic's customization
        self.selected_pairs       = []            # List of plotted data
        # Variables to manage figure
        self.figure               = None         
        self.ax                   = None
        self.canvas               = None
        self.toolbar              = None
        self.plot_name            = plot_name

        self.init_ui()

    def init_ui(self):
        '''
        Construct the user interface for the plot control panel.

        Creates button layouts for plot interaction, including creation, customization,
        and manipulation of data cycles. Also builds a scrollable area to add and manage
        cycle pairs dynamically.
        '''
        main_layout = QVBoxLayout()
        self.setLayout(main_layout)

        #==========================================#
        # SECTION BUTTONS (Top navigation)         #
        #==========================================#

        self.btn_data_tab     = QPushButton("Data")
        self.btn_style_tab    = QPushButton("Plot style")
        self.btn_analysis_tab = QPushButton("Analysis")
        self.btn_help         = QPushButton("Help")

        for btn in [self.btn_data_tab, self.btn_style_tab, self.btn_analysis_tab]:
            btn.setCheckable(True)


        self.section_group = QButtonGroup(self)
        self.section_group.setExclusive(True)
        self.section_group.addButton(self.btn_data_tab, 0)
        self.section_group.addButton(self.btn_style_tab, 1)
        self.section_group.addButton(self.btn_analysis_tab, 2)

        tab_layout = QHBoxLayout()
        tab_layout.addWidget(self.btn_data_tab)
        tab_layout.addWidget(self.btn_style_tab)
        tab_layout.addWidget(self.btn_analysis_tab)
        tab_layout.addWidget(self.btn_help)
        tab_layout.addStretch()

        main_layout.addLayout(tab_layout)

        def show_help_dialog():
            help_text = (
                "Plot Control Guide:\n\n"
                "- Use the 'Data' tab to select data pairs, add/remove cycles, or toggle cycle visibility.\n"
                "- Use the 'Plot style' tab to customize the appearance of the plot (colors, markers, etc.).\n"
                "- Use the 'Analysis' tab to perform data analysis operations like normalization, loop closure, curve fitting, data flipping and correction.\n"
            )
            QMessageBox.information(self, "Plot Control Guide", help_text)

        self.btn_help.clicked.connect(show_help_dialog)

        #==========================================#
        # STACKED TOOL SECTIONS                    #
        #==========================================#

        self.tools_stack = QStackedWidget()

        # -------------------- DATA GROUP --------------------------

        data_group  = QGroupBox("Select data, add to or hide cycles:")
        data_layout = QHBoxLayout()
        data_group.setLayout(data_layout)

        btn_create = QPushButton("Create plot")
        btn_create.clicked.connect(self.plot)

        btn_add = QPushButton("Add cycle")
        btn_add.clicked.connect(lambda: [self.add_pair(), self.add_pair()])

        btn_remove = QPushButton("Remove last cycle")
        btn_remove.clicked.connect(self.remove_last_cycle)

        btn_toggle = QPushButton("Hide cycle")
        btn_toggle.clicked.connect(self.toggle_cycle_visibility)

        data_layout.addWidget(btn_create)
        data_layout.addWidget(btn_add)
        data_layout.addWidget(btn_remove)
        data_layout.addWidget(btn_toggle)
        data_layout.addStretch()

        # -------------------- STYLE GROUP -------------------------

        style_group  = QGroupBox("Create and customize plots")
        style_layout = QHBoxLayout()
        style_group.setLayout(style_layout)

        btn_custom = QPushButton("Customization")
        btn_custom.clicked.connect(self.customize_plot_style)

        btn_appearance = QPushButton("Appearance")
        btn_appearance.clicked.connect(self.customize_plot_appearance)

        style_layout.addWidget(btn_custom)
        style_layout.addWidget(btn_appearance)
        style_layout.addStretch()

        # -------------------- ANALYSIS GROUP ----------------------

        analysis_group  = QGroupBox("Data Analysis")
        analysis_layout = QHBoxLayout()
        analysis_group.setLayout(analysis_layout)

        btn_normalize = QPushButton("Normalize")
        btn_normalize.clicked.connect(self.normalize)

        btn_close = QPushButton("Close loop")
        btn_close.clicked.connect(self.close_loop)

        btn_correct = QPushButton("Correct loop")
        btn_correct.clicked.connect(self.correction)

        btn_fit = QPushButton("Curve Fitting")
        btn_fit.clicked.connect(self.curve_fitting)

        flip_button = QPushButton("Flip data")
        flip_menu   = QMenu(flip_button)

        flip_x_action      = QAction("Flip X axis", self)
        flip_y_action      = QAction("Flip Y axis", self)
        flip_branch_action = QAction("Flip branch", self)

        flip_x_action.triggered.connect(self.x_inversion)
        flip_y_action.triggered.connect(self.y_inversion)
        flip_branch_action.triggered.connect(self.revert_branch)

        flip_menu.addAction(flip_x_action)
        flip_menu.addAction(flip_y_action)
        flip_menu.addSeparator()
        flip_menu.addAction(flip_branch_action)

        flip_button.setMenu(flip_menu)

        analysis_layout.addWidget(btn_normalize)
        analysis_layout.addWidget(btn_close)
        analysis_layout.addWidget(btn_correct)
        analysis_layout.addWidget(btn_fit)
        analysis_layout.addWidget(flip_button)
        analysis_layout.addStretch()

        # Add groups to stack
        self.tools_stack.addWidget(data_group)
        self.tools_stack.addWidget(style_group)
        self.tools_stack.addWidget(analysis_group)

        self.tools_stack.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)

        main_layout.addWidget(self.tools_stack)

        # Connect section switching
        self.section_group.buttonClicked[int].connect(
            self.tools_stack.setCurrentIndex
        )

        self.btn_data_tab.setChecked(True)

        #==========================================#
        # SCROLL AREA (Expandable)                 #
        #==========================================#

        self.scroll_area = QScrollArea()
        self.scroll_area.setWidgetResizable(True)

        self.pair_container = QWidget()
        self.pair_layout    = QVBoxLayout()
        self.pair_container.setLayout(self.pair_layout)

        self.scroll_area.setWidget(self.pair_container)

        main_layout.addWidget(self.scroll_area, stretch=1)

        # Add initial pairs
        self.add_pair()
        self.add_pair()

        
    def add_pair(self, file_text=None, x_col=None, y_col=None):
        '''
        Add a new data pair selector to the interface.

        Each pair consists of a file selection dropdown and corresponding x/y column selectors.
        Optionally pre-selects values for restoring saved sessions.

        Parameters
        ----------
        file_text : str, optional
            The name of the file to be preselected.
        x_col : str, optional
            The x-axis column name to be preselected.
        y_col : str, optional
            The y-axis column name to be preselected.
        '''

        if len(self.selected_pairs) % 2 == 0:
            num   = len(self.selected_pairs) // 2 + 1
            title = QLabel(f"Cycle {num}")
            title.setStyleSheet("font-weight: bold; margin-top: 10px; margin-bottom: 5px;")
            self.pair_layout.addWidget(title)

        
        file_combo = QComboBox()
        file_combo.addItems([f"File {i + 1}" for i in range(len(self.app_instance.dataframes))])

        x_combo = QComboBox()
        y_combo = QComboBox()

        row = QHBoxLayout()

        def update_columns():
            index = file_combo.currentIndex()
            cols = list(self.app_instance.dataframes[index].columns)
            x_combo.clear()
            y_combo.clear()
            x_combo.addItems(cols)
            y_combo.addItems(cols)

            # Selection for loading previous session
            if x_col in cols:
                x_combo.setCurrentText(x_col)
            if y_col in cols:
                y_combo.setCurrentText(y_col)

        file_combo.currentIndexChanged.connect(update_columns)

        # Selection for loading previous session
        if file_text:
            file_combo.setCurrentText(file_text)

        update_columns()

        row.addWidget(QLabel("File:"))
        row.addWidget(file_combo)
        row.addWidget(QLabel("x:"))
        row.addWidget(x_combo)
        row.addWidget(QLabel("y:"))
        row.addWidget(y_combo)

        container = QWidget()
        container.setLayout(row)
        self.pair_layout.addWidget(container)
        self.selected_pairs.append((file_combo, x_combo, y_combo))

        if len(self.selected_pairs) % 2 == 0:
            separator = QFrame()
            separator.setFrameShape(QFrame.HLine)
            separator.setFrameShadow(QFrame.Sunken)
            separator.setStyleSheet("margin: 8px 0px;")
            self.pair_layout.addWidget(separator)
    
    def remove_last_cycle(self):
        ''' Remove the last added cycle (pair of file/x/y selectors) from the layout
        '''
        if not self.selected_pairs:
            return
        
        for _ in range(2):
            
            file_combo, _, _ = self.selected_pairs.pop()
            parent_widget = file_combo.parent()  # QWidget that contains the row
            self.pair_layout.removeWidget(parent_widget)
            parent_widget.deleteLater()


        count = self.pair_layout.count()
        if count >= 1:
            last_item = self.pair_layout.itemAt(count - 1).widget()
            if isinstance(last_item, QFrame):
                self.pair_layout.takeAt(count - 1)
                last_item.deleteLater()
                count -= 1

        if count >= 1:
            last_item = self.pair_layout.itemAt(count - 1).widget()
            if isinstance(last_item, QLabel) and last_item.text().startswith("Cycle"):
                self.pair_layout.takeAt(count - 1)
                last_item.deleteLater()



    def toggle_cycle_visibility(self):
        ''' Call function to toggle the visibility of the data
        '''
        cycle_visibility(self, self.number_plots, 
                         self.app_instance.figures_map,
                         self.plot_customizations)

    def plot(self):
        ''' Call function to plot data
        '''
        plot_data(self, self.app_instance)

    def customize_plot_style(self):
        ''' Call function to customizzzation of plots
        '''
        customize_plot_style(self, self.plot_customizations,
                             self.number_plots, self.app_instance.figures_map)
    
    def customize_plot_appearance(self):
        ''' Call function to customize the appearance of the plot window
        '''
        customize_plot_appearance(self)

    def curve_fitting(self):
        ''' Curve fitting window
        '''
        open_curve_fitting_window(self.app_instance, self)
       
    def normalize(self):
        ''' Call function to normalize data
        '''
        norm_dialog(self, self.app_instance)

    def close_loop(self):
        ''' Call function to close loop
        '''
        close_loop_dialog(self, self.app_instance)

    def x_inversion(self):
        ''' Call function to invert x axis
        '''
        inv_x_dialog(self, self.app_instance)

    def y_inversion(self):
        ''' Call function to invert y axis
        '''
        inv_y_dialog(self, self.app_instance)

    def revert_branch(self):
        ''' Call function to revert a branch of a cycle
        '''
        inv_single_branch_dialog(self, self.app_instance)
    
    def correction(self):
        ''' Call function to correct a hysteresis loop
        '''
        correct_hysteresis_loop(self.app_instance)


#==============================================================================================#
# Class to overwrite close event to remove discarded figures                                   #
#==============================================================================================#

class PlotSubWindow(QMdiSubWindow):
    ''' 
    Class to overwrite close event to remove discarded figures
    Associates a plot widget with a subwindow and handles cleanup upon closing by 
    removing internal references to the widget and its figure.
    '''

    def __init__(self, app_instance, plot_widget, plot_id):
        '''
        Parameters
        ----------
        app_instance : MainApp
            Main application instance containing the session data.
        plot_widget : QWidget
            The widget containing the plot to be displayed.
        plot_id : int
            current plot number
        '''
        super().__init__()
        self.app_instance = app_instance
        self.plot_widget  = plot_widget
        self.plot_id      = plot_id
        self.setWidget(plot_widget)
        self.setWindowTitle(f"Control - {plot_widget.plot_name}")
        self.resize(600, 300)

    def closeEvent(self, event):
        '''
        Handle the window close event and remove associated plot references.

        Removes the control widget and its corresponding figure
        from the application's internal mappings.
        Logs the removal of the figure.

        Parameters
        ----------
        event : QCloseEvent
            The close event triggered when the window is closed.
        '''
        # Remove the control widget
        if self.plot_id in self.app_instance.plot_widgets:
            del self.app_instance.plot_widgets[self.plot_id]
            del self.app_instance.plot_names[self.plot_id]

        # Remove the associated figure
        if self.plot_id in self.app_instance.figures_map:
            del self.app_instance.figures_map[self.plot_id]
            self.app_instance.logger.info(f"Figure {self.plot_id} removed from figures_map.")

        event.accept()
        
#==============================================================================================#
# Function that creates the plot with the chosen data                                          #
#==============================================================================================#

def plot_data(plot_window_instance, app_instance):
    '''
    Create the plot with the selected pairs using matplotlib.
   
    Parameters
    ----------
    plot_window_instance : PlotControlWidget
        Instance of the plot control widget containing the selected pairs.
    app_instance : MainApp
        Main application instance containing the session data.
    '''

    # Extracting data from the plot window instance
    selected_pairs      = plot_window_instance.selected_pairs
    number_plots        = plot_window_instance.number_plots
    plot_name           = plot_window_instance.plot_name    
    dataframes          = app_instance.dataframes
    plot_customizations = plot_window_instance.plot_customizations
    logger              = app_instance.logger

    # For Normalization or loop closure
    app_instance.refresh_shell_variables()
    
    # Create a figure
    if plot_window_instance.figure is None:
        fig = Figure(figsize=(10, 6))
        ax  = fig.add_subplot(111)

        # Save objects in the instance
        plot_window_instance.figure = fig
        plot_window_instance.ax     = ax

        app_instance.figures_map[number_plots] = (fig, ax)

        # Create canvas and show in sub-window
        canvas  = FigureCanvas(fig)
        toolbar = NavigationToolbar(canvas, plot_window_instance)

        # Create layout
        plot_area = QWidget()
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(toolbar)
        layout.addWidget(canvas)
        plot_area.setLayout(layout)

        # Save 
        plot_window_instance.canvas  = canvas
        plot_window_instance.toolbar = toolbar 

        # Sub-window
        sub = QMdiSubWindow()
        sub.setWindowTitle(f"Plot - {plot_name}")
        sub.setWidget(plot_area)
        sub.resize(800, 600)
        app_instance.mdi_area.addSubWindow(sub)
        sub.show()
        # Save for session loading
        app_instance.figure_subwindows[number_plots] = sub


    else:
        # Retrieve existing objects
        fig     = plot_window_instance.figure
        ax      = plot_window_instance.ax
        canvas  = plot_window_instance.canvas
        toolbar = plot_window_instance.toolbar

        # Clear for new plot
        ax.clear()

    try:

        X = []
        Y = []

        for df_choice, x_var, y_var in selected_pairs:
            df_idx = int(df_choice.currentText().split(" ")[1]) - 1 
            x_col = x_var.currentText()
            y_col = y_var.currentText()

            if not x_col or not y_col:
                QMessageBox.critical(None, "Error", "You must select all column pairs!")
                return

            X.append(dataframes[df_idx][x_col].astype(float).values)
            Y.append(dataframes[df_idx][y_col].astype(float).values)
            logger.info(f"Plot of: {x_col} vs {y_col}")

        if not plot_customizations:
            col = plt.cm.jet(np.linspace(0, 1, len(X)))
            for i in range(0, len(X), 2):
                ax.plot(X[i],   Y[i],   color=col[i], marker=".", label=f"Cycle {i//2 + 1}")
                ax.plot(X[i+1], Y[i+1], color=col[i], marker=".")


        else:
            for i, (x, y) in enumerate(zip(X, Y)):
                if i % 2 == 0:
                    line1, = ax.plot(x, y, label=f"Cycle {i // 2 + 1}")
                else:
                    line2, = ax.plot(x, y)

                try:
                    customization = plot_customizations.get(i // 2, {})

                    line1.set_color(customization.get("color", line1.get_color()))
                    line1.set_marker(customization.get("marker", line1.get_marker()))
                    line1.set_linestyle(customization.get("linestyle", line1.get_linestyle()))
                    line1.set_label(customization.get("label", f"Cycle {i // 2 + 1}"))

                    if i % 2 == 1:
                        line2.set_color(customization.get("color", line1.get_color()))
                        line2.set_marker(customization.get("marker", line1.get_marker()))
                        line2.set_linestyle(customization.get("linestyle", line1.get_linestyle()))
                        line2.set_label("_nolegend_")

                except Exception as e:
                    print(f"Error applying style: {e}")

        ax.set_xlabel("H [Oe]", fontsize=15)
        ax.set_ylabel(r"M/M$_{sat}$", fontsize=15)
        ax.legend()
        
        # Add horizontal line at y=0
        ax.axhline(y=0, color='gray', linestyle='--', linewidth=1)
        # Add vertical line at x=0
        ax.axvline(x=0, color='gray', linestyle='--', linewidth=1)
                        
        canvas.draw()

    except Exception as e:
        QMessageBox.critical(None, "Error", f"Error creating plot: {e}")

#==============================================================================================#
# Function to customize the style of the plot                                                  #
#==============================================================================================#

def customize_plot_style(parent_widget, plot_customizations, number_plots, figures_map):
    '''
    Opens a PyQt5 dialog to customize color, marker, and line style of a cycle in the plot.

    Parameters
    ----------
    parent_widget : QWidget
        parent PyQt5 window
    plot_customizations : dict
        dictionary to save users customizations
    number_plots : list
        list with one element, current plot number
    figures_map : dict
        dictionary to store all the matplotlib figures
    '''
    
    if parent_widget.figure is None:
        QMessageBox.critical(parent_widget, "Error", "No plot open! Create a plot first.")
        return

    fig, ax = figures_map[number_plots]

    lines = ax.lines

    # Remove grid 
    filtered_lines = []
    for line in lines:
        x_data, y_data = line.get_xdata(), line.get_ydata()
       
        if not (
            (all(y == 0 for y in y_data) and len(set(x_data)) > 1) or  # axhline(0)
            (all(x == 0 for x in x_data) and len(set(y_data)) > 1) or  # axvline(0)
            (line.get_gid() == "fit")
        ):
            filtered_lines.append(line)
            
    lines = filtered_lines

    if not lines:
        QMessageBox.critical(parent_widget, "Error", "No lines present in the graph!")
        return

    # === All possible customization options ===
    colors       = list(mcolors.TABLEAU_COLORS) + list(mcolors.CSS4_COLORS)
    markers_list = [m for m in markers.MarkerStyle.markers.keys() if isinstance(m, str) and len(m) == 1]
    linestyles   = list(mlines.Line2D.lineStyles.keys())

    # === Cycle names ===
    cycles = []
    label_to_index = {}
    for i in range(0, len(lines), 2):
        label = plot_customizations.get(i // 2, {}).get("label", f"Cycle {i // 2 + 1}")
        cycles.append(label)
        label_to_index[label] = i // 2

    # === Dialog ===
    dialog = QDialog(parent_widget)
    dialog.setWindowTitle("Customize Graphic Style")
    dialog.setFixedSize(400, 360)

    layout = QVBoxLayout(dialog)
    form_layout = QFormLayout()
    layout.addLayout(form_layout)

    # === Widgets ===
    cycle_combo = QComboBox()
    cycle_combo.addItems(cycles)

    color_combo = QComboBox()
    color_combo.addItems(colors)
    color_combo.setEditable(True)

    marker_combo = QComboBox()
    marker_combo.addItems(markers_list)
    marker_combo.setEditable(True)

    linestyle_combo = QComboBox()
    linestyle_combo.addItems(linestyles)
    linestyle_combo.setEditable(True)

    label_edit = QLineEdit()
    label_edit.setText(cycles[0])

    # === Add to form ===
    form_layout.addRow("Cycle:", cycle_combo)
    form_layout.addRow("Color:", color_combo)
    form_layout.addRow("Marker:", marker_combo)
    form_layout.addRow("Linestyle:", linestyle_combo)
    form_layout.addRow("Legend label:", label_edit)

    # === Apply button ===
    apply_button = QPushButton("Apply")
    layout.addWidget(apply_button)

    def apply_style():
        try:
            idx = label_to_index[cycle_combo.currentText()]
            line1 = lines[idx * 2]
            line2 = lines[idx * 2 + 1]

            color = color_combo.currentText()
            marker = marker_combo.currentText()
            linestyle = linestyle_combo.currentText()
            legend_label = label_edit.text() or cycle_combo.currentText()

            # Apply style to both lines
            for line in (line1, line2):
                line.set_color(color)
                line.set_marker(marker)
                line.set_linestyle(linestyle)

            line1.set_label(legend_label)
            line2.set_label("_nolegend_")
            
            # Save customization's 
            plot_customizations[idx] = {
                "color": color,
                "marker": marker,
                "linestyle": linestyle,
                "label": legend_label,
            }

            ax.legend()
            fig.canvas.draw_idle()
            dialog.accept()

        except Exception as e:
            QMessageBox.critical(dialog, "Error", f"Error applying style:\n{e}")

    apply_button.clicked.connect(apply_style)

    dialog.exec_()

def customize_plot_appearance(parent_widget):
    '''
    Function to customize the appearance of the plot (font sizes, minor ticks).

    Parameters
    ----------
    parent_widget : QWidget
        parent PyQt5 window
    '''
    if parent_widget.figure is None:
        QMessageBox.critical(parent_widget, "Error", "No plot open! Create a plot first.")
        return

    fig, ax = parent_widget.figure, parent_widget.ax

    dialog = QDialog(parent_widget)
    dialog.setWindowTitle("Plot Appearance")
    dialog.setFixedSize(350, 250)
    layout = QFormLayout(dialog)

    # Font sizes
    label_fontsize_edit  = QLineEdit(str(ax.xaxis.label.get_size()))
    tick_fontsize_edit   = QLineEdit(str(ax.xaxis.get_ticklabels()[0].get_size()) if ax.xaxis.get_ticklabels() else "10")
    legend_fontsize_edit = QLineEdit("10")

    # Minor ticks (safe check)
    minor_ticks_checkbox = QCheckBox("Show minor ticks")
    minor_ticks_checkbox.setChecked(any(tick.tick1line.get_visible() for tick in ax.xaxis.get_minor_ticks()))


    layout.addRow("Axis label fontsize:", label_fontsize_edit)
    layout.addRow("Tick label fontsize:", tick_fontsize_edit)
    layout.addRow("Legend fontsize:",     legend_fontsize_edit)
    layout.addRow(minor_ticks_checkbox)

    buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
    layout.addWidget(buttons)

    def apply_changes():
        try:
            label_fs  = float(label_fontsize_edit.text())
            tick_fs   = float(tick_fontsize_edit.text())
            legend_fs = float(legend_fontsize_edit.text())

            ax.xaxis.label.set_fontsize(label_fs)
            ax.yaxis.label.set_fontsize(label_fs)

            for label in ax.get_xticklabels() + ax.get_yticklabels():
                label.set_fontsize(tick_fs)

            # Legend update
            leg = ax.get_legend()
            if leg:
                for text in leg.get_texts():
                    text.set_fontsize(legend_fs)

            # Minor ticks
            if minor_ticks_checkbox.isChecked():
                ax.minorticks_on()
            else:
                ax.minorticks_off()

            parent_widget.canvas.draw_idle()
            dialog.accept()
        except Exception as e:
            QMessageBox.critical(dialog, "Error", str(e))

    buttons.accepted.connect(apply_changes)
    buttons.rejected.connect(dialog.reject)
    dialog.exec_()


#==============================================================================================#
# Function to hide a plotted cycle                                                             #
#==============================================================================================#

def cycle_visibility(parent_widget, number_plots, figures_map, plot_customizations):
    '''
    Opens a dialog to select which cycles to show/hide in the plot.

    Parameters
    ----------
    parent_widget : QWidget
        parent PyQt5 window
    number_plots : list
        list with one element, current plot number
    figures_map : dict
        dictionary to store all the matplotlib figures
    plot_customizations : dict
        dictionary with customizations (used to get labels)
    '''

    if parent_widget.figure is None:
        QMessageBox.critical(parent_widget, "Error", "No plot open! Create a plot first.")
        return

    fig, ax = figures_map[number_plots]
    lines = ax.lines

    # Remove grid
    filtered_lines = []
    for line in lines:
        x_data, y_data = line.get_xdata(), line.get_ydata()
        if not (
            (all(y == 0 for y in y_data) and len(set(x_data)) > 1) or  # axhline(0)
            (all(x == 0 for x in x_data) and len(set(y_data)) > 1) or  # axvline(0)
            (line.get_gid() == "fit")
        ):
            filtered_lines.append(line)

    lines = filtered_lines
    if not lines:
        QMessageBox.critical(parent_widget, "Error", "No lines present in the graph!")
        return

    # === Cycle labels ===
    cycles = []
    label_to_index = {}
    visibility_map = {}

    for i in range(0, len(lines), 2): 
        label = plot_customizations.get(i // 2, {}).get("label", f"Cycle {i // 2 + 1}")
        cycles.append(label)
        label_to_index[label] = i // 2
        visibility_map[label] = lines[i].get_visible()

    # === Dialog ===
    dialog = QDialog(parent_widget)
    dialog.setWindowTitle("Show/Hide Cycles")

    layout        = QVBoxLayout(dialog)
    scroll_area   = QScrollArea()
    scroll_widget = QWidget()
    scroll_layout = QVBoxLayout(scroll_widget)

    layout.addWidget(QLabel("Select data to display:"))

    checkboxes = {}

    for label in cycles:
        cb = QCheckBox(label)
        cb.setChecked(visibility_map[label])
        scroll_layout.addWidget(cb)
        checkboxes[label] = cb

    scroll_widget.setLayout(scroll_layout)
    scroll_area.setWidgetResizable(True)
    scroll_area.setWidget(scroll_widget)
    layout.addWidget(scroll_area)

    # Buttons
    buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
    layout.addWidget(buttons)

    def apply_visibility():
        try : 
            for label, cb in checkboxes.items():
                idx = label_to_index[label]
                visible = cb.isChecked()
                # To avoid some problems with fit's plot
                lines[idx * 2 + 1].set_visible(visible)
                lines[idx * 2].set_visible(visible)

        except Exception as e:
            QMessageBox.critical(dialog, "Error", f"Visibility setting error:\n{e}")
        

        # Recreate legend only for visible objects
        handles, labels = ax.get_legend_handles_labels()
        visible_handles_labels = [(h, l) for h, l in zip(handles, labels) if h.get_visible()]
        if visible_handles_labels:
            handles, labels = zip(*visible_handles_labels)
            ax.legend(handles, labels)
        else:
            ax.legend().remove()  # No visible line => remove legend

        fig.canvas.draw_idle()
        dialog.accept()

    buttons.accepted.connect(apply_visibility)
    buttons.rejected.connect(dialog.reject)

    dialog.exec_() 

#==============================================================================================#
# Curve fitting function                                                                       #
#==============================================================================================#

def open_curve_fitting_window(app_instance, plot_widget):
    '''
    Apre una finestra per configurare il fitting dei dati.
    
    Parameters
    ----------
    app_instance : MainApp
        Istanza principale dell'applicazione.
    plot_widget : PlotControlWidget
        Istanza della finestra di controllo del plot corrente.
    '''
    dataframes    = app_instance.dataframes
    fit_results   = app_instance.fit_results
    logger        = app_instance.logger

    if not dataframes:
        QMessageBox.critical(app_instance, "Error", "No data loaded!")
        return

    window = QWidget()
    window.setWindowTitle("Quick Curve Fitting")
    layout = QVBoxLayout(window)
    window.setLayout(layout)

    def show_help_dialog():
        help_text = (
            "The fit function must be a function of the variable 'x' and "
            "the parameter names must be specified in the appropriate field.\n\n"
            "To establish the range, just read the cursor on the graph, the values are at the top right.\n\n"
            "AS A REMINDER, the 'Up' branch is the one on the right unless the x-axis has been inverted; "
            "in that case it will be the one on the left.\n\n"
            "ACHTUNG: the function must be written in Python, so for example |x| is abs(x), x^2 is x**2, and all "
            "other functions must be written with np. in front (i.e. np.cos(x), np.exp(x)), except for special functions, "
            "for which you must use the name used by the scipy.special library (i.e. scipy.special.erf becomes erf)"
        )

        QMessageBox.information(window, "Fitting Guide", help_text)

    help_button = QPushButton("Help")
    help_button.clicked.connect(show_help_dialog)
    layout.addWidget(help_button, alignment=Qt.AlignLeft)

    # Left: selection
    selection_layout = QHBoxLayout()
    layout.addLayout(selection_layout)

    param_layout = QGridLayout()
    selection_layout.addLayout(param_layout)
 

    param_layout.addWidget(QLabel("Select the file:"), 0, 0)
    file_combo = QComboBox()
    file_combo.addItems([f"File {i+1}" for i in range(len(dataframes))])
    param_layout.addWidget(file_combo, 0, 1)

    param_layout.addWidget(QLabel("Column X:"), 1, 0)
    x_combo = QComboBox()
    param_layout.addWidget(x_combo, 1, 1)

    param_layout.addWidget(QLabel("Column Y:"), 2, 0)
    y_combo = QComboBox()
    param_layout.addWidget(y_combo, 2, 1)

    def update_columns():
        idx = file_combo.currentIndex()
        cols = list(dataframes[idx].columns)
        x_combo.clear()
        y_combo.clear()
        x_combo.addItems(cols)
        y_combo.addItems(cols)

    file_combo.currentIndexChanged.connect(update_columns)
    update_columns()

    param_layout.addWidget(QLabel("x_start:"), 3, 0)
    x_start_edit = QLineEdit("0")
    param_layout.addWidget(x_start_edit, 3, 1)

    param_layout.addWidget(QLabel("x_end:"), 4, 0)
    x_end_edit = QLineEdit("1")
    param_layout.addWidget(x_end_edit, 4, 1)

    param_layout.addWidget(QLabel("Parameter names (es. a,b):"), 5, 0)
    param_names_edit = QLineEdit("a,b")
    param_layout.addWidget(param_names_edit, 5, 1)

    param_layout.addWidget(QLabel("Initial values (es. 1,1):"), 6, 0)
    initial_params_edit = QLineEdit("1,1")
    param_layout.addWidget(initial_params_edit, 6, 1)

    param_layout.addWidget(QLabel("Fitting function (es. a*(x-b)):"), 7, 0)
    function_edit = QLineEdit("a*(x - b)")
    param_layout.addWidget(function_edit, 7, 1)

    output_box = QTextEdit()
    output_box.setReadOnly(True)
    output_box.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
    selection_layout.addWidget(output_box)

    def perform_fit():
        try:

            df_idx  = file_combo.currentIndex()
            df      = dataframes[df_idx]
            x_col   = x_combo.currentText()
            y_col   = y_combo.currentText()

            x_data  = df[x_col].astype(float).values
            y_data  = df[y_col].astype(float).values

            x_start = float(x_start_edit.text())
            x_end   = float(x_end_edit.text())
            mask    = (x_data >= x_start) & (x_data <= x_end)
            x_fit   = x_data[mask]
            y_fit   = y_data[mask]

            if len(x_fit) == 0:
                QMessageBox.warning(window, "Error", "No data in the selected range!")
                return

            param_names    = [p.strip() for p in param_names_edit.text().split(",")]
            initial_params = [float(p.strip()) for p in initial_params_edit.text().split(",")]

            func_code = f"lambda x, {', '.join(param_names)}: {function_edit.text()}"
            fit_func  = eval(func_code)

            params, pcov = curve_fit(fit_func, x_fit, y_fit, p0=initial_params)
            y_plot = fit_func(np.linspace(x_start, x_end, 500), *params)

            fig = plot_widget.figure
            ax  = plot_widget.ax
            fit_line, = ax.plot(np.linspace(x_start, x_end, 500), y_plot, linestyle="--", color="green")
            fit_line.set_gid("fit")
            plot_widget.canvas.draw()

            result_lines = []
            for p, val, err in zip(param_names, params, np.sqrt(np.diag(pcov))):
                try:
                    result_lines.append(f"{p} = {format_value_error(val, err)}")    
                except Exception as e:
                    result_lines.append(f"{p} = {val:.6f} ± {err:.6f}")
                fit_results[p] = val
                fit_results[f"error_{p}"] = err

            for i , pi in zip(range(len(params)), param_names):
                for j , pj in zip(range(i+1, len(params)), param_names[i+1:]):
                    corr_ij = pcov[i, j]/np.sqrt(pcov[i, i]*pcov[j, j])
                    result_lines.append(f"corr({pi}, {pj}) = {corr_ij:.3f}")

            result = "\n".join(result_lines)
            output_box.setPlainText(result)
            logger.info("Fit completed successfully.")
            logger.info(f"Fitting function: {function_edit.text()}")
            logger.info(f"Fitting on data from File {df_idx + 1}, x: {x_col}, y: {y_col}, range: [{x_start}, {x_end}]")
            # Explicit cast to avoid newline issues in log file
            logger.info(f"The fit brought the following results: {str(result).replace(chr(10), ' ')}.")
            app_instance.refresh_shell_variables()

        except Exception as e:
            QMessageBox.critical(window, "Error", f"Error in fitting: {e}")
           
    fit_button = QPushButton("Run Fit")
    fit_button.clicked.connect(perform_fit)
    param_layout.addWidget(fit_button, 8, 0, 1, 2, alignment=Qt.AlignCenter)

    # Sub-window for fitting panel
    sub = QMdiSubWindow()
    sub.setWidget(window)
    sub.setWindowTitle("Quick Curve Fitting")
    sub.resize(600, 300)
    app_instance.mdi_area.addSubWindow(sub)
    sub.show()


import pytest
import numpy as np
import pandas as pd
from unittest.mock import MagicMock, patch

from hyloa.data.anisotropy import *

#==================================================================#
# Auxiliar function for mock all inputs                            #
#==================================================================#

@pytest.fixture
def mock_combo():
    combo = MagicMock()
    combo.currentIndex.return_value = 0
    combo.currentText.return_value = "H"
    return combo

@pytest.fixture
def mock_save_combo_no_save():
    combo = MagicMock()
    combo.currentIndex.return_value = 0  # No save
    return combo

@pytest.fixture
def mock_save_combo_file1():
    combo = MagicMock()
    combo.currentIndex.return_value = 1  # File 1
    return combo

@pytest.fixture
def mock_lineedit():
    le = MagicMock()
    le.text.return_value = "0.0"
    return le

@pytest.fixture
def mock_logger():
    return MagicMock()


@pytest.fixture
def mock_window():
    return MagicMock()


@pytest.fixture
def mock_draw_plot():
    return MagicMock()

@pytest.fixture
def mock_thr_edit():
    edit = MagicMock()
    edit.text.return_value = "0.05"
    return edit

@pytest.fixture
def mock_window():
    return MagicMock()


@pytest.fixture
def mock_output_box():
    box = MagicMock()
    return box


@pytest.fixture
def valid_plot_state():
    x    = np.linspace(-10, 10, 50)
    y_up = np.tanh(x)
    y_dw = -np.tanh(x)

    return {
        "x_up": x,
        "y_up": y_up,
        "x_dw": x,
        "y_dw": y_dw,
        "x_up_corr": x,
        "y_up_corr": y_up,
        "x_dw_corr": x,
        "y_dw_corr": y_dw,
        "e_up": np.full_like(x, 0.01),
        "e_dw": np.full_like(x, 0.01),
        "fit_hc_n": None,
        "fit_hc_p": None,
        "done_spl3": False,
        "spline_up": None,
        "spline_dw": None,
    }

@pytest.fixture
def plot_state_with_splines():
    x = np.linspace(-10, 10, 5000)

    # Up / down branches with small difference near saturation
    y_up = np.tanh(x)
    y_dw = np.tanh(x) * 0.98

    return {
        "spline_up": (x, y_up),
        "spline_dw": (x, y_dw),
    }

@pytest.fixture
def plot_state_with_splines_and_corr():
    x = np.linspace(-10, 10, 500)
    y_up = np.tanh(x)
    y_dw = np.tanh(x) * 0.95

    return {
        "spline_up": (x, y_up),
        "spline_dw": (x, y_dw),
        "x_up_corr": x,
        "x_dw_corr": x,
        "y_up_corr": y_up,
        "y_dw_corr": y_dw,
        "x_up": x,
        "y_up": y_up,
        "x_dw": x,
        "y_dw": y_dw,
    }

@pytest.fixture
def dataframes():
    df1 = pd.DataFrame({
        "H": np.zeros(500),
        "M": np.zeros(500),
    })

    df2 = pd.DataFrame({
        "H": np.zeros(500),
        "M": np.zeros(500),
    })

    return [df1, df2]

#==================================================================#
# Tests                                                            #
#==================================================================#


def test_compute_b_spline_success(
    mock_combo, mock_lineedit, mock_logger, mock_window, mock_draw_plot, valid_plot_state):
    # Act
    compute_b_spline(
        mock_combo, mock_combo, mock_combo,
        mock_combo, mock_combo, mock_combo,
        mock_lineedit, mock_lineedit,
        valid_plot_state,
        mock_logger,
        mock_window,
        mock_draw_plot
    )

    # Assert
    assert valid_plot_state["done_spl3"] is True
    assert valid_plot_state["spline_up"] is not None
    assert valid_plot_state["spline_dw"] is not None

    x_up, y_up = valid_plot_state["spline_up"]
    x_dw, y_dw = valid_plot_state["spline_dw"]

    assert len(x_up) == 5000
    assert len(y_up) == 5000
    assert len(x_dw) == 5000
    assert len(y_dw) == 5000

    mock_draw_plot.assert_called_once()
    mock_logger.info.assert_called_once()


@patch("hyloa.data.correction.QMessageBox.critical")
def test_compute_b_spline_negative_smoothing(
    mock_msgbox, mock_combo, mock_lineedit, mock_logger, mock_window, mock_draw_plot,
    valid_plot_state):

    mock_lineedit.text.return_value = "-1.0"

    compute_b_spline(
        mock_combo, mock_combo, mock_combo,
        mock_combo, mock_combo, mock_combo,
        mock_lineedit, mock_lineedit,
        valid_plot_state,
        mock_logger,
        mock_window,
        mock_draw_plot
    )

    mock_msgbox.assert_called_once()
    mock_draw_plot.assert_not_called()
    assert valid_plot_state["done_spl3"] is False


@patch("hyloa.data.correction.QMessageBox.critical")
def test_compute_b_spline_without_correction(
    mock_msgbox, mock_combo, mock_lineedit, mock_logger, mock_window, mock_draw_plot):
    plot_state = {
        "x_up_corr": None,
        "y_up_corr": None,
        "x_dw_corr": None,
        "y_dw_corr": None,
        "e_up": None,
        "e_dw": None,
    }

    compute_b_spline(
        mock_combo, mock_combo, mock_combo,
        mock_combo, mock_combo, mock_combo,
        mock_lineedit, mock_lineedit,
        plot_state,
        mock_logger,
        mock_window,
        mock_draw_plot
    )

    mock_msgbox.assert_called_once()
    mock_draw_plot.assert_not_called()


def test_compute_hk_success(
    mock_combo, mock_thr_edit, mock_logger, mock_window, mock_output_box, plot_state_with_splines):

    compute_Hk(
        mock_combo, mock_combo, mock_combo, mock_combo, mock_combo,
        mock_thr_edit, plot_state_with_splines, mock_logger, mock_window,
        mock_output_box
    )

    # Output box updated
    mock_output_box.setPlainText.assert_called_once()
    text = mock_output_box.setPlainText.call_args[0][0]

    assert "Negative anisotropy field" in text
    assert "Positive anisotropy field" in text
    assert "Mean:" in text

    # Logger called
    mock_logger.info.assert_called_once()

@patch("hyloa.data.correction.QMessageBox.critical")
def test_compute_hk_missing_splines(
    mock_msgbox, mock_combo, mock_thr_edit, mock_logger, mock_window, mock_output_box):

    plot_state = {
        "spline_up": None,
        "spline_dw": None,
    }

    compute_Hk(
        mock_combo, mock_combo, mock_combo, mock_combo, mock_combo,
        mock_thr_edit, plot_state, mock_logger, mock_window,
        mock_output_box
    )

    mock_msgbox.assert_called_once()
    mock_output_box.setPlainText.assert_not_called()

@patch("hyloa.data.correction.QMessageBox.critical")
def test_compute_hk_invalid_threshold(
    mock_msgbox, mock_combo, mock_logger, mock_window, mock_output_box, plot_state_with_splines):
    
    bad_thr = MagicMock()
    bad_thr.text.return_value = "not_a_number"

    compute_Hk(
        mock_combo, mock_combo, mock_combo, mock_combo, mock_combo,
        bad_thr, plot_state_with_splines, mock_logger, mock_window,
        mock_output_box
    )

    mock_msgbox.assert_called_once()
    mock_output_box.setPlainText.assert_not_called()


def test_symmetrize_with_save(
    mock_combo, mock_save_combo_file1, mock_logger, mock_draw_plot,
    mock_window, plot_state_with_splines_and_corr, dataframes):

    # Dest columns
    x_up_dest = MagicMock()
    y_up_dest = MagicMock()
    x_dw_dest = MagicMock()
    y_dw_dest = MagicMock()

    x_up_dest.currentText.return_value = "H"
    y_up_dest.currentText.return_value = "M"
    x_dw_dest.currentText.return_value = "H"
    y_dw_dest.currentText.return_value = "M"

    symmetrize(
        mock_combo, mock_save_combo_file1,
        mock_combo, mock_combo, mock_combo, mock_combo, mock_combo,
        x_up_dest, y_up_dest, x_dw_dest, y_dw_dest,
        dataframes, mock_logger, plot_state_with_splines_and_corr,
        mock_draw_plot, mock_window
    )

    df_dest = dataframes[0]

    assert not np.all(df_dest["H"] == 0)
    assert not np.all(df_dest["M"] == 0)

    mock_logger.info.assert_called()



# This file is part of HYLOA - HYsteresis LOop Analyzer.
# Copyright (C) 2024 Francesco Zeno Costanzo

# HYLOA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# HYLOA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with HYLOA. If not, see <https://www.gnu.org/licenses/>.

"""
Code to handle worksheet dataframes to ensure comunication
between different worksheets
"""

class WsData:
    '''
    Run-time storage of dataframes for each worksheet, to ensure communication between different worksheets
    and allow to create a plot with columns from different worksheets.
    '''
    
    def __init__(self):
        '''
        Initialization of dataframe storage
        '''
        # {worksheet_name : pandas.DataFrame}
        self._datasets = {}

    def add(self, name, df):
        '''
        Frunction to add or update a dataframe

        Parameters
        ----------
        name : str
            Name of the worksheet
        df : pandas.DataFrame
            data stored in the worksheet
        '''

        self._datasets[name] = df
    
    def remove(self, name):
        '''
        Frunction to remove a .

        Parameters
        ----------
        name : str
            Name of the worksheet
        ''' 

        if name in self._datasets:
            del self._datasets[name]
    
    def get(self, name):
        '''
        Get dataframe by worksheet name.

        Parameters
        ----------
        name : str
            Name of the worksheet
        
        Return
        ------
        pandas.DataFrame in name worksheet
        '''

        return self._datasets.get(name)
    
    def get_all(self):
        '''
        Get all dataframe.

        Return
        ------
        dict, all dataframes in all worksheet
        '''

        return self._datasets
    
    def get_all_columns(self):
        '''
        Return flattened structure:
        {
            "ws1::colA": (ws1, colA),
            "ws2::colB": (ws2, colB)
        }
        '''
        flat = {}
        for ws_name, df in self._datasets.items():
            for col in df.columns:
                flat[f"{ws_name}::{col}"] = (ws_name, col)
        return flat




# This file is part of HYLOA - HYsteresis LOop Analyzer.
# Copyright (C) 2024 Francesco Zeno Costanzo

# HYLOA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# HYLOA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with HYLOA. If not, see <https://www.gnu.org/licenses/>.


from setuptools import setup, find_packages


setup(
    name="hyloa", 
    version="1.11.1",
    author="Francesco Zeno Costanzo",
    author_email="zenofrancesco99@gmail.com",
    description="A package for analyzing hysteresis loops",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Francesco-Zeno-Costanzo/hyloa",
    packages=find_packages(where="."),  # Search package in hyloa/
    package_dir={"": "."},  # Means that hyloa/ in the root
    include_package_data=True,
    install_requires=[
        "PyQt5",
        "numpy",
        "scipy",
        "pandas",
        "matplotlib",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10.12",
    entry_points={
        "console_scripts": [
            "hyloa=hyloa.main:main",
        ],
    },
)



# This file is part of HYLOA - HYsteresis LOop Analyzer.
# Copyright (C) 2024 Francesco Zeno Costanzo

# HYLOA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# HYLOA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with HYLOA. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from scipy.interpolate import splrep, BSpline
from scipy.interpolate import InterpolatedUnivariateSpline

from PyQt5.QtWidgets import QMessageBox

from hyloa.utils.err_format import format_value_error


def compute_b_spline(file_combo, x_up_combo, y_up_combo, x_down_combo, y_down_combo,
                     data_sel, smooth_up_edit, smooth_dw_edit,
                     plot_state, logger, window, draw_plot):
    
    '''
    Compute cubic B-spline interpolations for the corrected hysteresis loop branches.

    This function applies a cubic B-spline interpolation separately to the
    corrected up and down branches of a hysteresis loop. The spline is computed
    using the corrected data stored in plot_state and is intended as a smooth
    representation of the loop after tilt/drift correction.

    The spline can only be computed after the correction step has been performed,
    since it operates on the corrected arrays (x_up_corr, y_up_corr,
    x_dw_corr, y_dw_corr). The resulting splines are stored in plot_state
    and immediately visualized via the provided plotting callback.

    For numerical stability and consistency with SciPy requirements, the data
    are internally sorted to ensure a strictly monotonic x-axis. Duplicate x
    values are explicitly checked and rejected, as they would invalidate the
    spline computation.

    Parameters
    ----------
    file_combo : QComboBox
        Combo box used to select the source data file (used for logging purposes).
    x_up_combo, y_up_combo : QComboBox
        Combo boxes selecting the x and y columns corresponding to the up branch.
    x_down_combo, y_down_combo : QComboBox
        Combo boxes selecting the x and y columns corresponding to the down branch.
    data_sel : QComboBox
        Combo box to select data type (corrected or original).
    smooth_up_edit : QLineEdit
        Input field specifying the smoothing parameter s for the up branch spline.
        Must be a non-negative float.
    smooth_dw_edit : QLineEdit
        Input field specifying the smoothing parameter s for the down branch spline.
        Must be a non-negative float.
    plot_state : dict
        Dictionary storing the current plotting state.
    logger : logging.Logger
        Logger instance used to record spline computation details.
    window : QWidget
        Parent widget used to display error message boxes.
    draw_plot : callable
        Callback function responsible for redrawing the plot after the spline
        computation.

    Notes
    -----
    - The spline is computed using scipy.interpolate.splrep and evaluated
      on a dense, uniformly spaced grid spanning the corrected x-range.
    - The spline represents the full corrected curve.
    '''
    
    try :

        idx_src  = file_combo.currentIndex()
        selected = data_sel.currentText()
        x_up_col = x_up_combo.currentText()
        y_up_col = y_up_combo.currentText()
        x_dw_col = x_down_combo.currentText()
        y_dw_col = y_down_combo.currentText()
    
        s_up = float(smooth_up_edit.text())
        s_dw = float(smooth_dw_edit.text())

        if s_up < 0 or s_dw < 0:
            QMessageBox.critical(window, "Error", "Smoothing parameter must be non-negative.")
            return

        #Read data
        if selected == "Corrected":
            x_up = plot_state["x_up_corr"]
            y_up = plot_state["y_up_corr"]
            x_dw = plot_state["x_dw_corr"]
            y_dw = plot_state["y_dw_corr"]
            e_up = plot_state["e_up"]
            e_dw = plot_state["e_dw"]
        else:
            x_up = plot_state["x_up"]
            y_up = plot_state["y_up"]
            x_dw = plot_state["x_dw"]
            y_dw = plot_state["y_dw"]
            e_up = None
            e_dw = None

        if x_up is None:
            QMessageBox.critical(window, "Error", "Spline must be applied on corrected data.")
            return
        
        #=========================================================#
        # Up branch spline                                        #
        #=========================================================#

        # Ensure monotonic x
        idx = np.argsort(x_up)
        x_up, y_up = x_up[idx], y_up[idx]

        # Check duplicates
        if np.any(np.diff(x_up) == 0):
            QMessageBox.critical(window, "Error", "Duplicate x values detected in up branch.")
            return

        # Compute spline
        try:
            tck_up = splrep(x_up, y_up, s=s_up)
        except Exception as e:
            QMessageBox.critical(window, "Error", f"Spline fit up branch failed: {e}.")
            return
        
        try :
            x_dense_up = np.linspace(x_up.min(), x_up.max(), 5000)
            y_dense_up = BSpline(*tck_up)(x_dense_up)
        except Exception as e:
            QMessageBox.critical(window, "Error", f"Error in spline conputation for up branch: {e}.")
            return
        
        #=========================================================#
        # Dw branch spline                                        #
        #=========================================================#

        # Ensure monotonic x
        idx = np.argsort(x_dw)
        x_dw, y_dw = x_dw[idx], y_dw[idx]

        # Check duplicates
        if np.any(np.diff(x_dw) == 0):
            QMessageBox.critical(window, "Error", "Duplicate x values detected in dw branch.")
            return

        # Compute spline
        try:
            tck_dw = splrep(x_dw, y_dw, s=s_dw)
        except Exception as e:
            QMessageBox.critical(window, "Error", f"Spline fit dw branch failed: {e}.")
            return
        
        try :
            x_dense_dw = np.linspace(x_dw.min(), x_dw.max(), 5000)
            y_dense_dw = BSpline(*tck_dw)(x_dense_dw)
        except Exception as e:
            QMessageBox.critical(window, "Error", f"Error in spline conputation for dw branch: {e}.")
            return 

        plot_state.update({
            "fit_hc_p"  : None,
            "fit_hc_n"  : None,
            "done_spl3" : True,
            "spline_up" : (x_dense_up, y_dense_up),
            "spline_dw" : (x_dense_dw, y_dense_dw)
        })
        draw_plot()

        log_lines = []
        log_lines.append(f"Computed Bspline for data in file {idx_src + 1}, columns {x_up_col}/{y_up_col} and {x_dw_col}/{y_dw_col}.")
        log_lines.append(f"With smoting parameter {s_up} for up branch and {s_dw} for down branch.\n")
        logger.info(f"Bspline results" + "\n".join(log_lines))
    
    except Exception as e:
        QMessageBox.critical(window, "Error", f"Error during spline interpolation:\n{e}")
        return



def compute_Hk(file_combo, x_up_combo, y_up_combo, x_down_combo, y_down_combo,
               hk_thr_edit, plot_state, logger, window, output_box):
    
    '''
    Compute the anisotropy field Hk.

    This function estimates the anisotropy field using the cubic spline
    representations of the up and down branches previously computed.
    The method is based on the relative difference between the two branches:
    regions where the difference falls below a given threshold are used
    to identify the negative and positive anisotropy fields.

    The final anisotropy field is computed as the mean value between the
    negative and positive branches, with an associated uncertainty.

    Parameters
    ----------
    file_combo : QComboBox
        Combo box used to select the source data file.
    x_up_combo : QComboBox
        Combo box selecting the X column for the up branch.
    y_up_combo : QComboBox
        Combo box selecting the Y column for the up branch.
    x_down_combo : QComboBox
        Combo box selecting the X column for the down branch.
    y_down_combo : QComboBox
        Combo box selecting the Y column for the down branch.
    hk_thr_edit : QComboBox or QLineEdit
        Widget containing the threshold value used to select the region
        where the relative difference between up and down branches is below
        the chosen limit.
    plot_state : dict
        Dictionary storing the current plot state.
    logger : logging.Logger
        Logger instance used to record the computation results.
    window : QWidget
        Parent widget used to display error message boxes.
    output_box : QTextEdit
        Text box where the computed anisotropy field and related information
        are displayed.

    Notes
    -----
    This function requires that the cubic splines for both branches have
    already been computed. If the spline data are not available, the
    computation will fail.
    '''
    
    try :
        idx_src  = file_combo.currentIndex()
        x_up_col = x_up_combo.currentText()
        y_up_col = y_up_combo.currentText()
        x_dw_col = x_down_combo.currentText()
        y_dw_col = y_down_combo.currentText()

        x_up, y_up = plot_state["spline_up"]
        x_dw, y_dw = plot_state["spline_dw"] 
        
    
        diff  = abs( (y_up - y_dw)/y_up )
        field = x_up
        thr_y = np.where(diff == np.max(diff))[0][0]

        thr = float(hk_thr_edit.text())

        pts     = np.where(diff < thr)[0]
        field_a = field[pts] 

        field_n = field_a[field_a < field[thr_y]]
        field_p = field_a[field_a > field[thr_y]]

        field_p = np.sort(field_p)
        field_n = np.sort(field_n)

        Hk_1 = field_n.max()
        Hk_2 = field_p.min()
        
        Hk = 0.5*(-Hk_1 + Hk_2)
        dH = abs(0.5*( Hk_1 + Hk_2))

        results_text_lines = []
        results_text_lines.append(f"Negative anisotropy field: {Hk_1}")
        results_text_lines.append(f"Positive anisotropy field: {Hk_2}")
        results_text_lines.append(f"Mean: {format_value_error(Hk, dH)}")
        
        output_box.setPlainText("\n".join(results_text_lines))
        
        log_lines = []
        log_lines.append(f"Computed anisotropy field for data in file {idx_src + 1}, columns {x_up_col}/{y_up_col} and {x_dw_col}/{y_dw_col}.")
        
        logger.info("\n".join(log_lines) +"\n".join(results_text_lines))
      
    
    except Exception as e:
        QMessageBox.critical(window, "Error", f"Error during anisotropy field calculation:\n{e}")

def symmetrize(file_combo, save_file_combo,
               x_up_combo, y_up_combo, x_down_combo, y_down_combo, data_sel,
               x_up_dest,  y_up_dest,  x_dw_dest,    y_dw_dest,
               dataframes, logger, plot_state, draw_plot,
               window):
    '''
    Symmetrize a hysteresis loop using spline-interpolated up and down branches.

    This function performs a symmetrization of the hysteresis loop starting
    from the cubic spline representations of the corrected up and down
    branches. The symmetrized loop is constructed by averaging the field
    values and antisymmetrizing the magnetization, enforcing physical
    symmetry conditions.

    A linear spline is built from the averaged data and evaluated on the
    original corrected field arrays. A small random noise, estimated from
    the data, is added to mimic experimental uncertainty.

    Optionally, the symmetrized data can be saved into a user-selected
    destination data file.

    Parameters
    ----------
    file_combo : QComboBox
        Combo box used to select the source data file.
    save_file_combo : QComboBox
        Combo box used to select the destination file for saving the
        symmetrized data. If set to "No save", the data are not written.
    x_up_combo : QComboBox
        Combo box selecting the X column for the up branch.
    y_up_combo : QComboBox
        Combo box selecting the Y column for the up branch.
    x_down_combo : QComboBox
        Combo box selecting the X column for the down branch.
    y_down_combo : QComboBox
        Combo box selecting the Y column for the down branch.
    data_sel : QComboBox
        Combo box to select data type (corrected or original).
    x_up_dest : QComboBox
        Combo box selecting the destination X column for the symmetrized
        up branch.
    y_up_dest : QComboBox
        Combo box selecting the destination Y column for the symmetrized
        up branch.
    x_dw_dest : QComboBox
        Combo box selecting the destination X column for the symmetrized
        down branch.
    y_dw_dest : QComboBox
        Combo box selecting the destination Y column for the symmetrized
        down branch.
    dataframes : list of pandas.DataFrame
        List of dataframes loaded in the application.
    logger : logging.Logger
        Logger instance used to record spline computation details.
    plot_state : dict
        Dictionary storing the current plotting state.
    draw_plot : callable
        Callback function responsible for redrawing the plot.
    window : QWidget
        Parent widget used to display error message boxes.

    Notes
    -----
    This function requires that both the drift correction and spline
    interpolation steps have already been completed.
    '''
    try :
        idx_src  = file_combo.currentIndex()
        selected = data_sel.currentText()
        x_up_col = x_up_combo.currentText()
        y_up_col = y_up_combo.currentText()
        x_dw_col = x_down_combo.currentText()
        y_dw_col = y_down_combo.currentText()

        save_choice  = save_file_combo.currentIndex()  # 0 = No save, >0 => file index adjust
        
        if save_choice == 0:
            save_idx = None
        else:
            # save_file_combo items: ["No save", "File 1", "File 2"...]
            save_idx = save_choice - 1

        x_up, y_up = plot_state["spline_up"]
        x_dw, y_dw = plot_state["spline_dw"] 
        
        if selected == "Corrected":
            x_data_up = plot_state["x_up_corr"]
            x_data_dw = plot_state["x_dw_corr"]
            y_data_up = plot_state["y_up_corr"]
        else:
            x_data_up = plot_state["x_up"]
            x_data_dw = plot_state["x_dw"]
            y_data_up = plot_state["y_up"]

        dy_data_err = np.std(y_data_up[0:25])

        x_mean = (x_up + x_dw)/2
        y_mean = (y_up - y_dw[::-1])/2

        spl = InterpolatedUnivariateSpline(x_mean, y_mean, k=1)

        dy_err = (2*np.random.random(x_data_up.size) - 1) * dy_data_err

        x_new_up, x_new_dw = x_data_up, -x_data_dw 
        y_new_up, y_new_dw = spl(x_data_up) + dy_err, -spl(x_data_dw) + dy_err

        plot_state.update({
            "s_data_up" : (x_new_up, y_new_up),
            "s_data_dw" : (x_new_dw, y_new_dw)
        })
        draw_plot()


        log_lines = []
        log_lines.append(f"Symmetrizied loop in file {idx_src + 1}, columns {x_up_col}/{y_up_col} and {x_dw_col}/{y_dw_col}.")
        
        logger.info("\n".join(log_lines))

        # Save corrected columns if requested
        if save_idx is not None:
            df_dest = dataframes[save_idx]

            df_dest[x_up_dest.currentText()] = x_new_up
            df_dest[y_up_dest.currentText()] = y_new_up
            
            df_dest[x_dw_dest.currentText()] = x_new_dw
            df_dest[y_dw_dest.currentText()] = y_new_dw

            logger.info("Corrected columns saved to destination file.")
      
    
    except Exception as e:
        QMessageBox.critical(window, "Error", f"Error symmetrization:\n{e}")



# This file is part of HYLOA - HYsteresis LOop Analyzer.
# Copyright (C) 2024 Francesco Zeno Costanzo

# HYLOA is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# HYLOA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with HYLOA. If not, see <https://www.gnu.org/licenses/>.

"""
Simple script to bump the version of the package.
"""

import re
import sys
import subprocess
from pathlib import Path
from argparse import ArgumentParser

BASE_DIR       = Path(__file__).resolve().parent
PYPROJECT_PATH = BASE_DIR.parent / "pyproject.toml"
INIT_PATH      = BASE_DIR.parent / "hyloa" / "__init__.py"
SETUP_PATH     = BASE_DIR.parent / "setup.py"

INCREMENT_MODES = ('major', 'minor', 'patch')
VERSION_PATTERN = r'(\d+)\.(\d+)\.(\d+)'


def read_version_from_init():
    content = INIT_PATH.read_text()
    match = re.search(r'__version__\s*=\s*[\'"]' + VERSION_PATTERN + r'[\'"]', content)
    if not match:
        raise RuntimeError("Version not found in __init__.py")
    return '.'.join(match.groups())


def increment_version(version, mode):
    major, minor, patch = map(int, version.split('.'))
    if mode == 'major':
        return f"{major + 1}.0.0"
    elif mode == 'minor':
        return f"{major}.{minor + 1}.0"
    elif mode == 'patch':
        return f"{major}.{minor}.{patch + 1}"
    else:
        raise ValueError(f"Invalid mode: {mode}")


def update_file(file_path, regex_pattern, new_version, label='version'):
    content = file_path.read_text()
    new_content = re.sub(
        regex_pattern,
        lambda m: m.group(0).replace(m.group(2), new_version),
        content
    )
    file_path.write_text(new_content)
    print(f" Updated {label} in {file_path}")


def git_commit_and_push(new_version):
    try:
        subprocess.run(["git", "add", str(PYPROJECT_PATH), str(SETUP_PATH), str(INIT_PATH)], check=True)

        subprocess.run(["git", "commit", "-m", f"Bump version to {new_version}"], check=True)
        subprocess.run(["git", "push"], check=True)
        print("Changes committed and pushed to origin.")
    except subprocess.CalledProcessError as e:
        print("Git command failed:", e)
        sys.exit(1)

def git_create_and_push_tag(new_version):
    tag_name = f"v{new_version}"
    try:
        # Create tag
        subprocess.run(["git", "tag", "-a", tag_name, "-m", f"Release {tag_name}"], check=True)

        # Push only the tag
        subprocess.run(["git", "push", "origin", tag_name], check=True)

        print(f"Tag {tag_name} created and uploaded.")
    except subprocess.CalledProcessError as e:
        print("Git tag command failed:", e)
        sys.exit(1)



def main(mode):
    if mode not in INCREMENT_MODES:
        raise RuntimeError(f"Invalid mode: {mode}. Choose from {INCREMENT_MODES}")

    old_version = read_version_from_init()
    new_version = increment_version(old_version, mode)
    print(f"Bumping version from {old_version} to {new_version}")

    update_file(INIT_PATH,
                rf'(__version__\s*=\s*[\'"])({VERSION_PATTERN})([\'"])',
                new_version,
                label="__init__.py")

    update_file(SETUP_PATH,
                rf'(version\s*=\s*[\'"])({VERSION_PATTERN})([\'"])',
                new_version,
                label="setup.py")

    update_file(PYPROJECT_PATH,
                rf'(version\s*=\s*[\'"])({VERSION_PATTERN})([\'"])',
                new_version,
                label="pyproject.toml")
    

    git_commit_and_push(new_version)
    git_create_and_push_tag(new_version)
    print("Version bump complete.")


if __name__ == "__main__":
    parser = ArgumentParser(description="Bump version")
    parser.add_argument('mode', choices=INCREMENT_MODES, help='Version increment mode')
    args = parser.parse_args()
    main(args.mode)
